@@ -39,13 +39,13 @@ import (
 	"golang.org/x/tools/gopls/internal/progress"
 	"golang.org/x/tools/gopls/internal/protocol"
 	"golang.org/x/tools/gopls/internal/settings"
-	"golang.org/x/tools/gopls/internal/util/astutil"
 	"golang.org/x/tools/gopls/internal/util/bug"
 	"golang.org/x/tools/gopls/internal/util/frob"
 	"golang.org/x/tools/gopls/internal/util/moremaps"
 	"golang.org/x/tools/gopls/internal/util/persistent"
 	"golang.org/x/tools/gopls/internal/util/safetoken"
-	"golang.org/x/tools/internal/analysisinternal"
+	"golang.org/x/tools/internal/analysis/driverutil"
+	"golang.org/x/tools/internal/astutil"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/facts"
 )
@@ -291,8 +291,7 @@ func (s *Snapshot) Analyze(ctx context.Context, pkgs map[PackageID]*metadata.Pac
 				// Trailing space is intentional: some LSP clients strip newlines.
 				msg := fmt.Sprintf(`Indexed %d/%d packages. (Set "analysisProgressReporting" to false to disable notifications.)`,
 					completed, len(nodes))
-				pct := 100 * float64(completed) / float64(len(nodes))
-				wd.Report(ctx, msg, pct)
+				wd.Report(ctx, msg, float64(completed)/float64(len(nodes)))
 			}
 		}
 	}
@@ -775,7 +774,7 @@ filterErrors:
 		// Suppress type errors in files with parse errors
 		// as parser recovery can be quite lossy (#59888).
 		for _, p := range ppkg.CompiledGoFiles() {
-			if p.ParseErr != nil && astutil.NodeContains(p.File, typeError.Pos) {
+			if p.ParseErr != nil && astutil.NodeContainsPos(p.File, typeError.Pos) {
 				continue filterErrors
 			}
 		}
@@ -891,9 +890,7 @@ func (act *action) String() string {
 func execActions(ctx context.Context, actions []*action) {
 	var wg sync.WaitGroup
 	for _, act := range actions {
-		wg.Add(1)
-		go func() {
-			defer wg.Done()
+		wg.Go(func() {
 			act.once.Do(func() {
 				execActions(ctx, act.hdeps) // analyze "horizontal" dependencies
 				act.result, act.summary, act.err = act.exec(ctx)
@@ -910,7 +907,7 @@ func execActions(ctx context.Context, actions []*action) {
 			if act.summary == nil {
 				panic("nil action.summary (#60551)")
 			}
-		}()
+		})
 	}
 	wg.Wait()
 }
@@ -1039,7 +1036,7 @@ func (act *action) exec(ctx context.Context) (any, *actionSummary, error) {
 			// ValidateFixes allows a fix.End to be slightly beyond
 			// EOF to avoid spurious assertions when reporting
 			// fixes as the end of truncated files; see #71659.
-			if err := analysisinternal.ValidateFixes(apkg.pkg.FileSet(), analyzer, d.SuggestedFixes); err != nil {
+			if err := driverutil.ValidateFixes(apkg.pkg.FileSet(), analyzer, d.SuggestedFixes); err != nil {
 				bug.Reportf("invalid SuggestedFixes: %v", err)
 				d.SuggestedFixes = nil
 			}
@@ -1065,12 +1062,12 @@ func (act *action) exec(ctx context.Context) (any, *actionSummary, error) {
 		// Read file from snapshot, to ensure reads are consistent.
 		//
 		// TODO(adonovan): make the dependency analysis sound by
-		// incorporating these additional files into the the analysis
+		// incorporating these additional files into the analysis
 		// hash. This requires either (a) preemptively reading and
 		// hashing a potentially large number of mostly irrelevant
 		// files; or (b) some kind of dynamic dependency discovery
 		// system like used in Bazel for C++ headers. Neither entices.
-		if err := analysisinternal.CheckReadable(pass, filename); err != nil {
+		if err := driverutil.CheckReadable(pass, filename); err != nil {
 			return nil, err
 		}
 		h, err := act.fsource.ReadFile(ctx, protocol.URIFromPath(filename))
