@@ -17,8 +17,10 @@ package main
 import (
 	"bytes"
 	"encoding/json"
+	"errors"
 	"fmt"
 	"go/ast"
+	"go/doc/comment"
 	"go/token"
 	"go/types"
 	"maps"
@@ -34,7 +36,7 @@ import (
 	"time"
 	"unicode"
 
-	"golang.org/x/tools/go/ast/astutil"
+	goastutil "golang.org/x/tools/go/ast/astutil"
 	"golang.org/x/tools/go/packages"
 	"golang.org/x/tools/gopls/internal/cache"
 	"golang.org/x/tools/gopls/internal/doc"
@@ -42,7 +44,7 @@ import (
 	"golang.org/x/tools/gopls/internal/mod"
 	"golang.org/x/tools/gopls/internal/settings"
 	"golang.org/x/tools/gopls/internal/util/safetoken"
-	internalastutil "golang.org/x/tools/internal/astutil"
+	"golang.org/x/tools/internal/astutil"
 )
 
 func main() {
@@ -198,13 +200,12 @@ func loadOptions(category reflect.Value, optsType types.Object, pkg *packages.Pa
 
 	var opts []*doc.Option
 	optsStruct := optsType.Type().Underlying().(*types.Struct)
-	for i := 0; i < optsStruct.NumFields(); i++ {
+	for typesField := range optsStruct.Fields() {
 		// The types field gives us the type.
-		typesField := optsStruct.Field(i)
 
 		// If the field name ends with "Options", assume it is a struct with
 		// additional options and process it recursively.
-		if h := strings.TrimSuffix(typesField.Name(), "Options"); h != typesField.Name() {
+		if h, ok := strings.CutSuffix(typesField.Name(), "Options"); ok {
 			// Keep track of the parent structs.
 			if hierarchy != "" {
 				h = hierarchy + "." + h
@@ -216,7 +217,7 @@ func loadOptions(category reflect.Value, optsType types.Object, pkg *packages.Pa
 			opts = append(opts, options...)
 			continue
 		}
-		path, _ := astutil.PathEnclosingInterval(file, typesField.Pos(), typesField.Pos())
+		path, _ := goastutil.PathEnclosingInterval(file, typesField.Pos(), typesField.Pos())
 		if len(path) < 2 {
 			return nil, fmt.Errorf("could not find AST node for field %v", typesField)
 		}
@@ -226,7 +227,7 @@ func loadOptions(category reflect.Value, optsType types.Object, pkg *packages.Pa
 		if !ok {
 			return nil, fmt.Errorf("unexpected AST path %v", path)
 		}
-		description, deprecation := astField.Doc.Text(), internalastutil.Deprecation(astField.Doc)
+		description, deprecation := astField.Doc.Text(), astutil.Deprecation(astField.Doc)
 
 		// The reflect field gives us the default value.
 		reflectField := category.FieldByName(typesField.Name())
@@ -313,12 +314,12 @@ func loadEnums(pkg *packages.Package) (map[types.Type][]doc.EnumValue, error) {
 		if err != nil {
 			return nil, fmt.Errorf("finding file for %q: %v", cnst.Name(), err)
 		}
-		path, _ := astutil.PathEnclosingInterval(f, cnst.Pos(), cnst.Pos())
+		path, _ := goastutil.PathEnclosingInterval(f, cnst.Pos(), cnst.Pos())
 		spec := path[1].(*ast.ValueSpec)
 		value := cnst.Val().ExactString()
 		docstring := valueDoc(cnst.Name(), value, spec.Doc.Text())
 		var status string
-		for _, d := range internalastutil.Directives(spec.Doc) {
+		for _, d := range astutil.Directives(spec.Doc) {
 			if d.Tool == "gopls" && d.Name == "status" {
 				status = d.Args
 				break
@@ -465,7 +466,7 @@ func loadLenses(settingsPkg *packages.Package, defaults map[settings.CodeLensSou
 							return nil, fmt.Errorf("%s: %s lacks doc comment", posn, spec.Names[0].Name)
 						}
 						enumDoc[value] = spec.Doc.Text()
-						for _, d := range internalastutil.Directives(spec.Doc) {
+						for _, d := range astutil.Directives(spec.Doc) {
 							if d.Tool == "gopls" && d.Name == "status" {
 								enumStatus[value] = d.Args
 								break
@@ -500,9 +501,10 @@ func loadLenses(settingsPkg *packages.Package, defaults map[settings.CodeLensSou
 		}
 		return nil
 	}
-	addAll(golang.CodeLensSources(), "Go")
-	addAll(mod.CodeLensSources(), "go.mod")
-	return lenses, nil
+	err := errors.Join(
+		addAll(golang.CodeLensSources(), "Go"),
+		addAll(mod.CodeLensSources(), "go.mod"))
+	return lenses, err
 }
 
 func loadAnalyzers(analyzers []*settings.Analyzer, defaults *settings.Options) []*doc.Analyzer {
@@ -758,7 +760,23 @@ func rewriteAnalyzers(prevContent []byte, api *doc.API) ([]byte, error) {
 		title, doc, _ := strings.Cut(analyzer.Doc, "\n")
 		title = strings.TrimPrefix(title, analyzer.Name+": ")
 		fmt.Fprintf(&buf, "## `%s`: %s\n\n", analyzer.Name, title)
-		fmt.Fprintf(&buf, "%s\n\n", doc)
+
+		// Convert Analyzer.Doc from go/doc/comment form to Markdown.
+		// Headings in doc comments are converted to ### (HeadingLevel=3).
+		//
+		// Some Analyzers (e.g. go/analysis/passes/inline) use ## to indicate subheadings
+		// Although this is valid Markdown, it is not valid go/doc/comment,
+		// nor is it rendered as a subheading by pkg.go.dev or gopls's doc viewer.
+		// Perhaps it will be supported in future; see
+		// https://github.com/golang/go/issues/51082#issuecomment-1033116430 et seq.
+		//
+		// In the meantime, the go/doc/comment processing will escape them so
+		// that the ## appears literally in analyzers.md just as it does in
+		// the two viewers mentioned above. The meaning is clear enough.
+		doctree := new(comment.Parser).Parse(doc)
+		buf.Write((&comment.Printer{HeadingLevel: 3}).Markdown(doctree))
+		buf.WriteString("\n\n")
+
 		fmt.Fprintf(&buf, "Default: %s.", onOff(analyzer.Default))
 		if !analyzer.Default {
 			fmt.Fprintf(&buf, " Enable by setting `\"analyses\": {\"%s\": true}`.", analyzer.Name)
