@@ -14,7 +14,7 @@ package golang
 // - rewrite using html/template.
 //   Or factor with golang.org/x/pkgsite/internal/godoc/dochtml.
 // - emit breadcrumbs for parent + sibling packages.
-// - list promoted methods---we have type information!
+// - list promoted methods---we have type information! (golang/go#67158)
 // - gather Example tests, following go/doc and pkgsite.
 // - add option for doc.AllDecls: show non-exported symbols too.
 // - style the <li> bullets in the index as invisible.
@@ -43,13 +43,12 @@ import (
 	"slices"
 	"strings"
 
-	"golang.org/x/tools/go/ast/astutil"
+	goastutil "golang.org/x/tools/go/ast/astutil"
 	"golang.org/x/tools/gopls/internal/cache"
 	"golang.org/x/tools/gopls/internal/cache/parsego"
 	"golang.org/x/tools/gopls/internal/protocol"
-	goplsastutil "golang.org/x/tools/gopls/internal/util/astutil"
 	"golang.org/x/tools/gopls/internal/util/bug"
-	"golang.org/x/tools/gopls/internal/util/safetoken"
+	"golang.org/x/tools/internal/astutil"
 	"golang.org/x/tools/internal/stdlib"
 	"golang.org/x/tools/internal/typesinternal"
 )
@@ -79,7 +78,7 @@ func DocFragment(pkg *cache.Package, pgf *parsego.File, start, end token.Pos) (p
 		// External test packages don't have /pkg doc pages,
 		// so instead show the doc for the package under test.
 		// (This named-based heuristic is imperfect.)
-		if forTest := strings.TrimSuffix(pkg.Path(), "_test"); forTest != pkg.Path() {
+		if forTest, ok := strings.CutSuffix(pkg.Path(), "_test"); ok {
 			return PackagePath(forTest), "", makeTitle("package", nil, filepath.Base(forTest))
 		}
 
@@ -193,7 +192,7 @@ type thing struct {
 }
 
 func thingAtPoint(pkg *cache.Package, pgf *parsego.File, start, end token.Pos) thing {
-	path, _ := astutil.PathEnclosingInterval(pgf.File, start, end)
+	path, _ := goastutil.PathEnclosingInterval(pgf.File, start, end)
 
 	// In an import spec?
 	if len(path) >= 3 { // [...ImportSpec GenDecl File]
@@ -328,7 +327,17 @@ func PackageDocHTML(viewID string, pkg *cache.Package, web Web) ([]byte, error)
 			filterValues(&t.Vars)
 			filterFuncs(&t.Funcs)
 			filterFuncs(&t.Methods)
-			return unexported(t.Name)
+			if unexported(t.Name) {
+				// If an unexported type has an exported constructor function,
+				// treat the constructor as an ordinary standalone function.
+				// We will sort Funcs again below.
+				docpkg.Funcs = append(docpkg.Funcs, t.Funcs...)
+				return true // delete this type
+			}
+			return false // keep this type
+		})
+		slices.SortFunc(docpkg.Funcs, func(x, y *doc.Func) int {
+			return strings.Compare(x.Name, y.Name)
 		})
 	}
 
@@ -557,7 +566,7 @@ window.addEventListener('load', function() {
 		// appear as separate decls. We should too.
 		var buf bytes.Buffer
 		for _, file := range pkg.CompiledGoFiles() {
-			if goplsastutil.NodeContains(file.File, n.Pos()) {
+			if astutil.NodeContainsPos(file.File, n.Pos()) {
 				pos := n.Pos()
 
 				// emit emits source in the interval [pos:to] and updates pos.
@@ -567,15 +576,12 @@ window.addEventListener('load', function() {
 					if !to.IsValid() {
 						bug.Reportf("invalid Pos")
 					}
-					start, err := safetoken.Offset(file.Tok, pos)
+					text, err := file.PosText(pos, to)
 					if err != nil {
-						bug.Reportf("invalid start Pos: %v", err)
+						bug.Reportf("invalid pos range: %v", err)
+						return
 					}
-					end, err := safetoken.Offset(file.Tok, to)
-					if err != nil {
-						bug.Reportf("invalid end Pos: %v", err)
-					}
-					buf.WriteString(escape(string(file.Src[start:end])))
+					buf.WriteString(escape(string(text)))
 					pos = to
 				}
 				ast.Inspect(n, func(n ast.Node) bool {
