@@ -9,9 +9,9 @@ import (
 	"strings"
 	"testing"
 
-	"golang.org/x/tools/gopls/internal/lsp/protocol"
-	. "golang.org/x/tools/gopls/internal/test/integration"
+	"golang.org/x/tools/gopls/internal/protocol"
 	"golang.org/x/tools/gopls/internal/test/compare"
+	. "golang.org/x/tools/gopls/internal/test/integration"
 )
 
 func TestPrepareRenameMainPackage(t *testing.T) {
@@ -181,8 +181,23 @@ func main() {
 		env.Rename(env.RegexpSearch("lib/a.go", "lib"), "nested")
 
 		// Check if the new package name exists.
+		// Don't rename subpackages.
 		env.RegexpSearch("nested/a.go", "package nested")
 		env.RegexpSearch("main.go", `nested2 "mod.com/nested"`)
+		env.RegexpSearch("main.go", "mod.com/nested")
+		env.RegexpSearch("main.go", "mod.com/lib/nested")
+		env.RegexpSearch("main.go", `nested1 "mod.com/lib/x"`)
+	})
+
+	WithOptions(Settings{"renameMovesSubpackages": true}).Run(t, files, func(t *testing.T, env *Env) {
+		env.OpenFile("lib/a.go")
+		env.Rename(env.RegexpSearch("lib/a.go", "lib"), "nested")
+
+		// Check if the new package name exists.
+		// Rename subpackages.
+		env.RegexpSearch("nested/a.go", "package nested")
+		env.RegexpSearch("main.go", `nested2 "mod.com/nested"`)
+		env.RegexpSearch("main.go", "mod.com/nested")
 		env.RegexpSearch("main.go", "mod.com/nested/nested")
 		env.RegexpSearch("main.go", `nested1 "mod.com/nested/x"`)
 	})
@@ -223,7 +238,7 @@ func main() {
 		// Check if the new package name exists.
 		env.RegexpSearch("nested/a.go", "package nested")
 		env.RegexpSearch("main.go", "mod.com/nested")
-		env.RegexpSearch("main.go", `lib1 "mod.com/nested/nested"`)
+		env.RegexpSearch("main.go", `lib1 "mod.com/lib/nested"`)
 	})
 }
 
@@ -262,6 +277,18 @@ func main() {
 		// Check if the new package name exists.
 		env.RegexpSearch("nested/a.go", "package nested")
 		env.RegexpSearch("main.go", "mod.com/nested")
+		// Don't rename subpackages.
+		env.RegexpSearch("main.go", `foo "mod.com/lib/nested"`)
+	})
+
+	WithOptions(Settings{"renameMovesSubpackages": true}).Run(t, files, func(t *testing.T, env *Env) {
+		env.OpenFile("lib/a.go")
+		env.Rename(env.RegexpSearch("lib/a.go", "lib"), "nested")
+
+		// Check if the new package name exists.
+		env.RegexpSearch("nested/a.go", "package nested")
+		env.RegexpSearch("main.go", "mod.com/nested")
+		// Rename subpackages.
 		env.RegexpSearch("main.go", `foo "mod.com/nested/nested"`)
 	})
 }
@@ -307,7 +334,7 @@ func main() {
 		env.RegexpSearch("lib1/a.go", "package lib1")
 		env.RegexpSearch("lib1/b.go", "package lib1")
 		env.RegexpSearch("main.go", "mod.com/lib1")
-		env.RegexpSearch("main.go", "mod.com/lib1/nested")
+		env.RegexpSearch("main.go", "mod.com/lib/nested")
 	})
 }
 
@@ -477,7 +504,7 @@ func main() {
 		env.RegexpSearch("lib1/a.go", "package lib1")
 		env.RegexpSearch("lib1/b.go", "package lib1")
 		env.RegexpSearch("main.go", "mod.com/lib1")
-		env.RegexpSearch("main.go", "mod.com/lib1/nested")
+		env.RegexpSearch("main.go", "mod.com/lib/nested")
 
 		// Check if the test package is renamed
 		env.RegexpSearch("lib1/a_test.go", "package lib1_test")
@@ -554,6 +581,24 @@ func main() {
 		env.Rename(env.RegexpSearch("foo/foo.go", "foo"), "foox")
 
 		env.RegexpSearch("foox/foo.go", "package foox")
+		// Don't rename subpackages.
+		env.OpenFile("foo/bar/bar.go")
+		env.OpenFile("foo/bar/go.mod")
+
+		env.RegexpSearch("main.go", "mod.com/foo/bar")
+		env.RegexpSearch("main.go", "mod.com/foox")
+		env.RegexpSearch("main.go", "foox.Bar()")
+
+		env.RegexpSearch("go.mod", "./foo/bar")
+		env.RegexpSearch("go.mod", "./foo/baz")
+	})
+
+	WithOptions(Settings{"renameMovesSubpackages": true}).Run(t, files, func(t *testing.T, env *Env) {
+		env.OpenFile("foo/foo.go")
+		env.Rename(env.RegexpSearch("foo/foo.go", "foo"), "foox")
+
+		env.RegexpSearch("foox/foo.go", "package foox")
+		// Rename subpackages.
 		env.OpenFile("foox/bar/bar.go")
 		env.OpenFile("foox/bar/go.mod")
 
@@ -603,6 +648,19 @@ func main() {
 		env.RegexpSearch("nested/a.go", "package nested")
 		env.RegexpSearch("main.go", "mod.com/nested")
 		env.RegexpSearch("main.go", `lib1 "mod.com/nested"`)
+		// Don't rename subpackages.
+		env.RegexpSearch("main.go", `lib2 "mod.com/lib/nested"`)
+	})
+
+	WithOptions(Settings{"renameMovesSubpackages": true}).Run(t, files, func(t *testing.T, env *Env) {
+		env.OpenFile("lib/a.go")
+		env.Rename(env.RegexpSearch("lib/a.go", "lib"), "nested")
+
+		// Check if the new package name exists.
+		env.RegexpSearch("nested/a.go", "package nested")
+		env.RegexpSearch("main.go", "mod.com/nested")
+		env.RegexpSearch("main.go", `lib1 "mod.com/nested"`)
+		// Rename subpackages.
 		env.RegexpSearch("main.go", `lib2 "mod.com/nested/nested"`)
 	})
 }
@@ -644,6 +702,19 @@ func main() {
 		env.RegexpSearch("nested/a.go", "package nested")
 		env.RegexpSearch("main.go", "mod.com/nested")
 		env.RegexpSearch("main.go", `_ "mod.com/nested"`)
+		// Don't rename subpackages.
+		env.RegexpSearch("main.go", `lib1 "mod.com/lib/nested"`)
+	})
+
+	WithOptions(Settings{"renameMovesSubpackages": true}).Run(t, files, func(t *testing.T, env *Env) {
+		env.OpenFile("lib/a.go")
+		env.Rename(env.RegexpSearch("lib/a.go", "lib"), "nested")
+
+		// Check if the new package name exists.
+		env.RegexpSearch("nested/a.go", "package nested")
+		env.RegexpSearch("main.go", "mod.com/nested")
+		env.RegexpSearch("main.go", `_ "mod.com/nested"`)
+		// Rename subpackages.
 		env.RegexpSearch("main.go", `lib1 "mod.com/nested/nested"`)
 	})
 }
@@ -799,7 +870,7 @@ const C = lib.A + nested.B
 -- testdata/libx/a.go --
 package libx
 
-import "mod.com/libx/nested"
+import "mod.com/lib/nested"
 
 const A = 1 + nested.B
 -- testdata/other/other.go --
@@ -807,7 +878,7 @@ package other
 
 import (
 	"mod.com/libx"
-	"mod.com/libx/nested"
+	"mod.com/lib/nested"
 )
 
 const C = libx.A + nested.B
@@ -840,7 +911,7 @@ const A = 1 + nested.B
 
 		for _, badName := range []string{"$$$", "lib_test"} {
 			if err := env.Editor.Rename(env.Ctx, loc, badName); err == nil {
-				t.Errorf("Rename(lib, libx) succeeded, want non-nil error")
+				t.Errorf("Rename(lib, libx) succeeded unexpectedly")
 			}
 		}
 	})
@@ -896,11 +967,45 @@ func main() {
 
 		// Check if the new package name exists.
 		env.RegexpSearch("lib1/a.go", "package lib1")
-		env.RegexpSearch("lib1/a.go", "mod.com/lib1/internal/utils")
+		// Don't rename subpackages.
+		env.RegexpSearch("lib1/a.go", "mod.com/lib/internal/utils")
 		env.RegexpSearch("main.go", `import "mod.com/lib1"`)
 		env.RegexpSearch("main.go", "lib1.print()")
 	})
 }
+func TestRenamePackage_InvalidPackageMove(t *testing.T) {
+	const files = `
+-- go.mod --
+module mod.com
+
+go 1.21
+-- test1/test1.go --
+package test1
+
+-- test2/test2.go --
+package test2
+
+`
+	Run(t, files, func(t *testing.T, env *Env) {
+		env.OpenFile("test1/test1.go")
+		loc := env.RegexpSearch("test1/test1.go", "test1")
+		for _, test := range []struct{ badName, expectedErrMatch string }{
+			{"test2", "invalid package identifier"},
+			{"mod.com/test2", "not empty"},
+			{"differentmod.com/test2", "cannot move package across module boundary"},
+			{"mod.com/$$$", "invalid package name"},
+			{"mod*.$+", "invalid package path"},
+			{"mod.commmm", "cannot move package across module boundary"},
+		} {
+			err := env.Editor.Rename(env.Ctx, loc, test.badName)
+			if err == nil {
+				t.Errorf("Rename to %s succeeded unexpectedly", test.badName)
+			} else if !strings.Contains(err.Error(), test.expectedErrMatch) {
+				t.Errorf("Rename to %s produced incorrect error message, got %s, want %s", test.badName, err.Error(), test.expectedErrMatch)
+			}
+		}
+	})
+}
 
 // checkTestdata checks that current buffer contents match their corresponding
 // expected content in the testdata directory.
