@@ -194,7 +194,23 @@ func Test(t *testing.T) {
 				diags:      make(map[protocol.Location][]protocol.Diagnostic),
 				extraNotes: make(map[protocol.DocumentURI]map[string][]*expect.Note),
 			}
-			defer run.env.Shutdown()
+			defer run.env.Close()
+
+			// Support built-in pseudo-locations here.
+			// fmtLoc coerces "got" Locations to these forms
+			// so they can be compared without depending
+			// on line numbers in external files.
+			//
+			// (We could actually define "builtin:int" etc
+			// with the correct file/line/col for each
+			// builtin symbol, but that doesn't seem to be
+			// where the bugs are.)
+			run.values["BUILTIN"] = protocol.Location{
+				URI: "file:///pseudo/builtin/builtin.go",
+			}
+			run.values["UNSAFE"] = protocol.Location{
+				URI: "file:///pseudo/unsafe/unsafe.go",
+			}
 
 			// Open all files so that we operate consistently with LSP clients, and
 			// (pragmatically) so that we have a Mapper available via the fake
@@ -297,7 +313,7 @@ func Test(t *testing.T) {
 				t.Errorf("formatTest: %v", err)
 			} else if *update {
 				filename := filepath.Join(dir, test.name)
-				if err := os.WriteFile(filename, formatted, 0644); err != nil {
+				if err := os.WriteFile(filename, formatted, 0o644); err != nil {
 					t.Error(err)
 				}
 			} else if !t.Failed() {
@@ -361,14 +377,6 @@ func (mark marker) mapper() *protocol.Mapper {
 	return mapper
 }
 
-// error reports an error with a prefix indicating the position of the marker
-// note.
-func (mark marker) error(args ...any) {
-	mark.T().Helper()
-	msg := fmt.Sprint(args...)
-	mark.T().Errorf("%s: %s", mark.run.fmtPos(mark.note.Pos), msg)
-}
-
 // errorf reports a formatted error with a prefix indicating the position of
 // the marker note.
 //
@@ -426,7 +434,7 @@ func valueMarkerFunc(fn any) func(marker) {
 		args := append([]any{mark}, mark.note.Args[1:]...)
 		argValues, err := convertArgs(mark, ftype, args)
 		if err != nil {
-			mark.error(err)
+			mark.errorf("%v", err)
 			return
 		}
 		results := reflect.ValueOf(fn).Call(argValues)
@@ -469,7 +477,7 @@ func actionMarkerFunc(fn any, allowedNames ...string) func(marker) {
 		args := append([]any{mark}, mark.note.Args...)
 		argValues, err := convertArgs(mark, ftype, args)
 		if err != nil {
-			mark.error(err)
+			mark.errorf("%v", err)
 			return
 		}
 		reflect.ValueOf(fn).Call(argValues)
@@ -580,7 +588,7 @@ var valueMarkerFuncs = map[string]func(marker){
 // See doc.go for marker documentation.
 var actionMarkerFuncs = map[string]func(marker){
 	"acceptcompletion": actionMarkerFunc(acceptCompletionMarker),
-	"codeaction":       actionMarkerFunc(codeActionMarker, "end", "result", "edit", "err"),
+	"codeaction":       actionMarkerFunc(codeActionMarker, "end", "diag", "action", "result", "edit", "err"),
 	"codelenses":       actionMarkerFunc(codeLensesMarker),
 	"complete":         actionMarkerFunc(completeMarker),
 	"def":              actionMarkerFunc(defMarker),
@@ -612,7 +620,7 @@ var actionMarkerFuncs = map[string]func(marker){
 	"token":            actionMarkerFunc(tokenMarker),
 	"typedef":          actionMarkerFunc(typedefMarker),
 	"workspacesymbol":  actionMarkerFunc(workspaceSymbolMarker),
-	"mcptool":          actionMarkerFunc(mcpToolMarker, "output"),
+	"mcptool":          actionMarkerFunc(mcpToolMarker, "location", "output"),
 }
 
 // markerTest holds all the test data extracted from a test txtar archive.
@@ -765,7 +773,7 @@ func loadMarkerTests(dir string) ([]*markerTest, error) {
 				return err
 			}
 
-			name := strings.TrimPrefix(path, dir+string(filepath.Separator))
+			name := filepath.ToSlash(strings.TrimPrefix(path, dir+string(filepath.Separator)))
 			test, err := loadMarkerTest(name, content)
 			if err != nil {
 				return fmt.Errorf("%s: %v", path, err)
@@ -987,15 +995,13 @@ func newEnv(t *testing.T, cache *cache.Cache, files, proxyFiles map[string][]byt
 
 	awaiter := integration.NewAwaiter(sandbox.Workdir)
 
-	var eventChan chan lsprpc.SessionEvent
+	ss := lsprpc.NewStreamServer(cache, false, nil)
+
 	var mcpServer *httptest.Server
 	if enableMCP {
-		eventChan = make(chan lsprpc.SessionEvent)
-		mcpServer = httptest.NewServer(internalmcp.HTTPHandler(eventChan, cache, false))
+		mcpServer = httptest.NewServer(internalmcp.HTTPHandler(ss, false))
 	}
 
-	ss := lsprpc.NewStreamServer(cache, false, eventChan, nil)
-
 	server := servertest.NewPipeServer(ss, jsonrpc2.NewRawStream)
 	editor, err := fake.NewEditor(sandbox, config).Connect(ctx, server, awaiter.Hooks())
 	if err != nil {
@@ -1014,7 +1020,7 @@ func newEnv(t *testing.T, cache *cache.Cache, files, proxyFiles map[string][]byt
 	var mcpSession *mcp.ClientSession
 	if enableMCP {
 		client := mcp.NewClient("test", "v1.0.0", nil)
-		mcpSession, err = client.Connect(ctx, mcp.NewSSEClientTransport(mcpServer.URL))
+		mcpSession, err = client.Connect(ctx, mcp.NewSSEClientTransport(mcpServer.URL, nil))
 		if err != nil {
 			t.Fatalf("fail to connect to mcp server: %v", err)
 		}
@@ -1028,7 +1034,6 @@ func newEnv(t *testing.T, cache *cache.Cache, files, proxyFiles map[string][]byt
 		Awaiter:    awaiter,
 		MCPSession: mcpSession,
 		MCPServer:  mcpServer,
-		EventChan:  eventChan,
 	}
 }
 
@@ -1100,6 +1105,16 @@ func (run *markerTestRun) fmtLoc(loc protocol.Location) string {
 		run.env.TB.Errorf("unable to find %s in test archive", loc)
 		return "<invalid location>"
 	}
+
+	// Format builtin locations as "file:///pseudo/builtin/builtin.go:0:0",
+	// which is the predefined BUILTIN marker; ditto UNSAFE.
+	if strings.HasSuffix(string(loc.URI), "src/unsafe/unsafe.go") ||
+		strings.HasSuffix(string(loc.URI), "src/builtin/builtin.go") {
+		name := filepath.Base(filepath.Dir(loc.URI.Path())) // "unsafe" or "builtin"
+		loc.URI = protocol.DocumentURI("file:///pseudo/" + name + "/" + name + ".go")
+		loc.Range = protocol.Range{}
+	}
+
 	lines := bytes.Count(run.test.archive.Comment, []byte("\n"))
 	var name string
 	for _, f := range run.test.archive.Files {
@@ -1186,9 +1201,9 @@ func summarizeLoc(name string, startLine, startCol, endLine, endCol int) string
 
 // Types with special handling.
 var (
-	goldenType        = reflect.TypeOf(&Golden{})
-	markerType        = reflect.TypeOf(marker{})
-	stringMatcherType = reflect.TypeOf(stringMatcher{})
+	goldenType        = reflect.TypeFor[*Golden]()
+	markerType        = reflect.TypeFor[marker]()
+	stringMatcherType = reflect.TypeFor[stringMatcher]()
 )
 
 // Custom conversions.
@@ -1198,8 +1213,8 @@ var (
 //
 // Converters should return an error rather than calling marker.errorf().
 var customConverters = map[reflect.Type]func(marker, any) (any, error){
-	reflect.TypeOf(protocol.Location{}): converter(convertLocation),
-	reflect.TypeOf(completionLabel("")): converter(convertCompletionLabel),
+	reflect.TypeFor[protocol.Location](): converter(convertLocation),
+	reflect.TypeFor[completionLabel]():   converter(convertCompletionLabel),
 }
 
 // converter transforms a typed argument conversion function to an untyped
@@ -1253,9 +1268,8 @@ func convertNamedArgLocation(mark marker, arg any) (protocol.Location, error) {
 		if v, ok := mark.run.values[id]; ok {
 			if loc, ok := v.(protocol.Location); ok {
 				return loc, nil
-			} else {
-				return protocol.Location{}, fmt.Errorf("invalid location value %v", v)
 			}
+			return protocol.Location{}, fmt.Errorf("invalid location value %v", v)
 		}
 	}
 	return convertLocation(mark, arg)
@@ -1414,7 +1428,6 @@ func (sm stringMatcher) check(mark marker, got string) {
 		if !sm.pattern.MatchString(got) {
 			mark.errorf("got %q, does not match pattern %#q", got, sm.pattern)
 		}
-
 	} else if !strings.Contains(got, sm.substr) {
 		// Content must contain the expected substring.
 		mark.errorf("got %q, want substring %q", got, sm.substr)
@@ -1428,9 +1441,8 @@ func checkChangedFiles(mark marker, changed map[string][]byte, golden *Golden) {
 		if want, ok := golden.Get(mark.T(), filename, got); !ok {
 			mark.errorf("%s: unexpected change to file %s; got:\n%s",
 				mark.note.Name, filename, got)
-
 		} else if string(got) != string(want) {
-			mark.errorf("%s: wrong file content for %s: got:\n%s\nwant:\n%s\ndiff:\n%s",
+			mark.errorf("%s: wrong file content for %s:\n-- got --\n%s\n-- want --\n%s\n-- diff --\n%s",
 				mark.note.Name, filename, got, want,
 				compare.Bytes(want, got))
 		}
@@ -1481,7 +1493,6 @@ func checkDiffs(mark marker, changed map[string][]byte, golden *Golden) {
 		if want, ok := golden.Get(mark.T(), filename, []byte(got)); !ok {
 			mark.errorf("%s: unexpected change to file %s; got diff:\n%s",
 				mark.note.Name, filename, got)
-
 		} else if got != string(want) {
 			mark.errorf("%s: wrong diff for %s:\n\ngot:\n%s\n\nwant:\n%s\n",
 				mark.note.Name, filename, got, want)
@@ -1680,7 +1691,6 @@ func acceptCompletionMarker(mark marker, src protocol.Location, label string, go
 	filename := mark.path()
 	mapper := mark.mapper()
 	patched, _, err := protocol.ApplyEdits(mapper, append([]protocol.TextEdit{edit}, selected.AdditionalTextEdits...))
-
 	if err != nil {
 		mark.errorf("ApplyProtocolEdits failed: %v", err)
 		return
@@ -1705,11 +1715,16 @@ func defMarker(mark marker, loc protocol.Location, want ...protocol.Location) {
 	}
 }
 
-func typedefMarker(mark marker, src, dst protocol.Location) {
-	got := mark.run.env.FirstTypeDefinition(src)
-	if got != dst {
-		mark.errorf("type definition location does not match:\n\tgot: %s\n\twant %s",
-			mark.run.fmtLoc(got), mark.run.fmtLoc(dst))
+func typedefMarker(mark marker, loc protocol.Location, want ...protocol.Location) {
+	env := mark.run.env
+	got, err := env.Editor.TypeDefinitions(env.Ctx, loc)
+	if err != nil {
+		mark.errorf("typeDefinition request failed: %v", err)
+		return
+	}
+
+	if err := compareLocations(mark, got, want); err != nil {
+		mark.errorf("typedef failed: %v", err)
 	}
 }
 
@@ -1799,6 +1814,7 @@ func highlightLocationMarker(mark marker, loc protocol.Location, kindName expect
 		Kind:  kind,
 	}
 }
+
 func sortDocumentHighlights(s []protocol.DocumentHighlight) {
 	sort.Slice(s, func(i, j int) bool {
 		return protocol.CompareRange(s[i].Range, s[j].Range) < 0
@@ -1847,6 +1863,15 @@ func hoverMarker(mark marker, src, dst protocol.Location, sc stringMatcher) {
 	gotMD := ""
 	if content != nil {
 		gotMD = content.Value
+
+		// Normalize the absolute path for cross-platform consistency before
+		// replacement. This ensures Windows paths (e.g., "C:/...") become
+		// "/C:/..." to match Unix-like paths.
+		abs := mark.run.env.Sandbox.Workdir.RootURI().Path()
+		if protocol.IsWindowsDrivePath(abs) {
+			abs = "/" + abs
+		}
+		gotMD = strings.ReplaceAll(gotMD, filepath.ToSlash(abs), "$WORKDIR")
 	}
 	sc.check(mark, gotMD)
 }
@@ -2025,7 +2050,7 @@ func rename(env *integration.Env, loc protocol.Location, newName string) (map[st
 // See also:
 //   - Editor.applyWorkspaceEdit ../integration/fake/editor.go for the
 //     implementation of this operation used in normal testing.
-//   - cmdClient.applyWorkspaceEdit in ../../../cmd/cmd.go for the
+//   - client.applyWorkspaceEdit in ../../../cmd/cmd.go for the
 //     CLI variant.
 func changedFiles(env *integration.Env, changes []protocol.DocumentChange) (map[string][]byte, error) {
 	uriToPath := env.Sandbox.Workdir.URIToPath
@@ -2072,17 +2097,43 @@ func changedFiles(env *integration.Env, changes []protocol.DocumentChange) (map[
 
 		case change.RenameFile != nil:
 			old := change.RenameFile.OldURI
-			m, err := read(old)
-			if err != nil {
-				return nil, err // missing
-			}
-			write(old, nil)
-
 			new := change.RenameFile.NewURI
-			if _, err := read(old); err == nil {
-				return nil, fmt.Errorf("RenameFile: destination %s exists", new)
+			info, err := os.Stat(old.Path())
+			if err != nil {
+				return nil, err
+			}
+			if info.IsDir() {
+				// Walk through all the files in the old directory and copy
+				// their content to the new directory.
+				// TODO(mkalil): This currently only handles renaming the file's
+				// innermost directory. Need to handle renames of outer directories
+				// directories when implementing package move refactoring.
+				for _, file := range env.ListFiles(old.Path()) {
+					oldFile := protocol.URIFromPath(path.Join(old.Path(), path.Base(file)))
+					m, err := read(oldFile)
+					if err != nil {
+						return nil, err // missing
+					}
+					write(oldFile, nil)
+
+					newFile := protocol.URIFromPath(path.Join(new.Path(), path.Base(file)))
+					if _, err := read(oldFile); err == nil {
+						return nil, fmt.Errorf("RenameFile: destination %s exists", new)
+					}
+					write(newFile, m.Content)
+				}
+			} else {
+				m, err := read(old)
+				if err != nil {
+					return nil, err // missing
+				}
+				write(old, nil)
+
+				if _, err := read(old); err == nil {
+					return nil, fmt.Errorf("RenameFile: destination %s exists", new)
+				}
+				write(new, m.Content)
 			}
-			write(new, m.Content)
 
 		case change.CreateFile != nil:
 			uri := change.CreateFile.URI
@@ -2094,7 +2145,22 @@ func changedFiles(env *integration.Env, changes []protocol.DocumentChange) (map[
 		case change.DeleteFile != nil:
 			uri := change.DeleteFile.URI
 			if _, err := read(uri); err != nil {
-				return nil, fmt.Errorf("DeleteFile %s: file does not exist", uri)
+				// The specified URI could be a file or a directory.
+				files := env.ListFiles(uriToPath(uri))
+				if files == nil {
+					return nil, fmt.Errorf("DeleteFile %s: file does not exist", uri)
+				}
+				if len(files) > 0 {
+					// Directory is not empty. If "Recursive" is true, we delete
+					// every file in the folder. Otherwise, we continue without
+					// deleting the directory.
+					if change.DeleteFile.Options != nil && change.DeleteFile.Options.Recursive {
+						for _, f := range files {
+							write(protocol.URIFromPath(f), nil)
+						}
+					}
+					continue
+				}
 			}
 			write(uri, nil)
 
@@ -2117,26 +2183,74 @@ func changedFiles(env *integration.Env, changes []protocol.DocumentChange) (map[
 }
 
 func codeActionMarker(mark marker, loc protocol.Location, kind string) {
-	if !exactlyOneNamedArg(mark, "edit", "result", "err") {
+	if !exactlyOneNamedArg(mark, "action", "edit", "result", "err") {
 		return
 	}
 
 	if end := namedArgFunc(mark, "end", convertNamedArgLocation, protocol.Location{}); end.URI != "" {
 		if end.URI != loc.URI {
-			panic("unreachable")
+			mark.errorf("end marker is in a different file (%s)", filepath.Base(loc.URI.Path()))
+			return
 		}
 		loc.Range.End = end.Range.End
 	}
 
 	var (
-		edit    = namedArg(mark, "edit", expect.Identifier(""))
-		result  = namedArg(mark, "result", expect.Identifier(""))
-		wantErr = namedArgFunc(mark, "err", convertStringMatcher, stringMatcher{})
+		edit       = namedArg(mark, "edit", expect.Identifier(""))
+		result     = namedArg(mark, "result", expect.Identifier(""))
+		wantAction = namedArg(mark, "action", expect.Identifier(""))
+		wantErr    = namedArgFunc(mark, "err", convertStringMatcher, stringMatcher{})
 	)
 
-	changed, err := codeAction(mark.run.env, loc.URI, loc.Range, protocol.CodeActionKind(kind), nil)
-	if err != nil && wantErr.empty() {
-		mark.errorf("codeAction failed: %v", err)
+	var diag *protocol.Diagnostic
+	if re := namedArg(mark, "diag", (*regexp.Regexp)(nil)); re != nil {
+		d, ok := removeDiagnostic(mark, loc, false, re)
+		if !ok {
+			mark.errorf("no diagnostic at %v matches %q", loc, re)
+			return
+		}
+		diag = &d
+	}
+
+	action, err := resolveCodeAction(mark.run.env, loc.URI, loc.Range, protocol.CodeActionKind(kind), diag)
+	if err != nil {
+		if !wantErr.empty() {
+			wantErr.checkErr(mark, err)
+		} else {
+			mark.errorf("resolveCodeAction failed: %v", err)
+		}
+		return
+	}
+
+	// If when 'action' is set, we simply compare the action, and don't apply it.
+	if wantAction != "" {
+		g := mark.getGolden(wantAction)
+		if action == nil {
+			mark.errorf("no matching codeAction")
+			return
+		}
+		data, err := json.MarshalIndent(action, "", "\t")
+		if err != nil {
+			mark.errorf("failed to marshal codeaction: %v", err)
+			return
+		}
+		data = bytes.ReplaceAll(data, []byte(mark.run.env.Sandbox.Workdir.RootURI()), []byte("$WORKDIR"))
+		compareGolden(mark, data, g)
+		return
+	}
+
+	changes, err := applyCodeAction(mark.run.env, action)
+	if err != nil {
+		if !wantErr.empty() {
+			wantErr.checkErr(mark, err)
+		} else {
+			mark.errorf("codeAction failed: %v", err)
+		}
+		return
+	}
+	changed, err := changedFiles(mark.run.env, changes)
+	if err != nil {
+		mark.errorf("changedFiles failed: %v", err)
 		return
 	}
 
@@ -2267,29 +2381,20 @@ func quickfixErrMarker(mark marker, loc protocol.Location, re *regexp.Regexp, wa
 // applied. Currently, this function does not support code actions that return
 // edits directly; it only supports code action commands.
 func codeAction(env *integration.Env, uri protocol.DocumentURI, rng protocol.Range, kind protocol.CodeActionKind, diag *protocol.Diagnostic) (map[string][]byte, error) {
-	changes, err := codeActionChanges(env, uri, rng, kind, diag)
+	action, err := resolveCodeAction(env, uri, rng, kind, diag)
+	if err != nil {
+		return nil, err
+	}
+	changes, err := applyCodeAction(env, action)
 	if err != nil {
 		return nil, err
 	}
 	return changedFiles(env, changes)
 }
 
-// codeActionChanges executes a textDocument/codeAction request for the
-// specified location and kind, and captures the resulting document changes.
-// If diag is non-nil, it is used as the code action context.
-func codeActionChanges(env *integration.Env, uri protocol.DocumentURI, rng protocol.Range, kind protocol.CodeActionKind, diag *protocol.Diagnostic) ([]protocol.DocumentChange, error) {
-	// Collect any server-initiated changes created by workspace/applyEdit.
-	//
-	// We set up this handler immediately, not right before executing the code
-	// action command, so we can assert that neither the codeAction request nor
-	// codeAction resolve request cause edits as a side effect (golang/go#71405).
-	var changes []protocol.DocumentChange
-	restore := env.Editor.Client().SetApplyEditHandler(func(ctx context.Context, wsedit *protocol.WorkspaceEdit) error {
-		changes = append(changes, wsedit.DocumentChanges...)
-		return nil
-	})
-	defer restore()
-
+// resolveCodeAction resolves the code action specified by the given location,
+// kind, and diagnostic.
+func resolveCodeAction(env *integration.Env, uri protocol.DocumentURI, rng protocol.Range, kind protocol.CodeActionKind, diag *protocol.Diagnostic) (*protocol.CodeAction, error) {
 	// Request all code actions that apply to the diagnostic.
 	// A production client would set Only=[kind],
 	// but we can give a better error if we don't filter.
@@ -2327,16 +2432,6 @@ func codeActionChanges(env *integration.Env, uri protocol.DocumentURI, rng proto
 	}
 	action := candidates[0]
 
-	// Apply the codeAction.
-	//
-	// Spec:
-	//  "If a code action provides an edit and a command, first the edit is
-	//  executed and then the command."
-	// An action may specify an edit and/or a command, to be
-	// applied in that order. But since applyDocumentChanges(env,
-	// action.Edit.DocumentChanges) doesn't compose, for now we
-	// assert that actions return one or the other.
-
 	// Resolve code action edits first if the client has resolve support
 	// and the code action has no edits.
 	if action.Edit == nil {
@@ -2349,10 +2444,41 @@ func codeActionChanges(env *integration.Env, uri protocol.DocumentURI, rng proto
 			if err != nil {
 				return nil, err
 			}
-			action.Edit = resolved.Edit
+			action = *resolved
 		}
 	}
 
+	return &action, nil
+}
+
+// applyCodeAction applies the given code action, and captures the resulting
+// document changes.
+func applyCodeAction(env *integration.Env, action *protocol.CodeAction) ([]protocol.DocumentChange, error) {
+	// Collect any server-initiated changes created by workspace/applyEdit.
+	//
+	// We set up this handler immediately, not right before executing the code
+	// action command, so we can assert that neither the codeAction request nor
+	// codeAction resolve request cause edits as a side effect (golang/go#71405).
+	var changes []protocol.DocumentChange
+	restore := env.Editor.Client().SetApplyEditHandler(func(ctx context.Context, wsedit *protocol.WorkspaceEdit) error {
+		changes = append(changes, wsedit.DocumentChanges...)
+		return nil
+	})
+	defer restore()
+
+	if action.Edit == nil && action.Command == nil {
+		panic("bad action")
+	}
+
+	// Apply the codeAction.
+	//
+	// Spec:
+	//  "If a code action provides an edit and a command, first the edit is
+	//  executed and then the command."
+	// An action may specify an edit and/or a command, to be
+	// applied in that order. But since applyDocumentChanges(env,
+	// action.Edit.DocumentChanges) doesn't compose, for now we
+	// assert that actions return one or the other.
 	if action.Edit != nil {
 		if len(action.Edit.Changes) > 0 {
 			env.TB.Errorf("internal error: discarding unexpected CodeAction{Kind=%s, Title=%q}.Edit.Changes", action.Kind, action.Title)
@@ -2442,17 +2568,45 @@ func implementationMarker(mark marker, src protocol.Location, want ...protocol.L
 	}
 }
 
-func mcpToolMarker(mark marker, tool string, rawArgs string, loc protocol.Location) {
+func mcpToolMarker(mark marker, tool string, rawArgs string) {
+	if !mark.run.test.mcp {
+		mark.errorf("mcp not enabled: add -mcp")
+		return
+	}
 	args := make(map[string]any)
 	if err := json.Unmarshal([]byte(rawArgs), &args); err != nil {
 		mark.errorf("fail to unmarshal arguments to map[string]any: %v", err)
 		return
 	}
 
-	// TODO(hxjiang): Make the "location" key configurable.
-	args["location"] = loc
+	// substitutePaths replaces instances of $WORKDIR in string or []string values with
+	// the actual working directory.
+	var substitutePaths func(any) any
+	substitutePaths = func(v any) any {
+		switch v := v.(type) {
+		case string:
+			return strings.ReplaceAll(v, "$WORKDIR", mark.run.env.Sandbox.Workdir.RootURI().Path())
 
-	res, err := mcp.CallTool(mark.ctx(), mark.run.env.MCPSession, &mcp.CallToolParams[map[string]any]{
+		case []any:
+			for i, e := range v {
+				if _, ok := e.(string); ok {
+					v[i] = substitutePaths(e).(string)
+				}
+			}
+		}
+		return v
+	}
+	// Hack: replace '$WORKDIR' in string arguments with the actual sandbox
+	// working directory.
+	for k, v := range args {
+		args[k] = substitutePaths(v)
+	}
+
+	if loc := namedArg(mark, "location", protocol.Location{}); loc != (protocol.Location{}) {
+		args["location"] = loc
+	}
+
+	res, err := mark.run.env.MCPSession.CallTool(mark.ctx(), &mcp.CallToolParams{
 		Name:      tool,
 		Arguments: args,
 	})
@@ -2474,13 +2628,18 @@ func mcpToolMarker(mark marker, tool string, rawArgs string, loc protocol.Locati
 	}
 
 	got := buf.String()
+	// For portability, replace all (potential) filepath separators with "/".
+	got = strings.ReplaceAll(got, string(filepath.Separator), "/")
+	// To ensure consistent unified diff output, the working directory path
+	// is replaced with "$WORKDIR". This addresses cases where MCP tools
+	// include absolute file paths in generated diffs.
+	got = strings.ReplaceAll(got, filepath.ToSlash(mark.run.env.Sandbox.Workdir.RootURI().Path()), "$WORKDIR")
 
 	output := namedArg(mark, "output", expect.Identifier(""))
 	golden := mark.getGolden(output)
-	if want, ok := golden.Get(mark.T(), "", []byte(got)); !ok {
-		mark.errorf("%s: missing golden file @%s", mark.note.Name, golden.id)
-	} else if diff := cmp.Diff(got, string(want)); diff != "" {
-		mark.errorf("unexpected mcp tools call %s return; got:\n%s\n want:\n%s\ndiff:\n%s", tool, got, want, diff)
+	want, _ := golden.Get(mark.T(), "", []byte(got))
+	if diff := compare.Text(string(want), got); diff != "" {
+		mark.errorf("unexpected mcp tools call %s return: diff:\n%s", tool, diff)
 	}
 }
 
@@ -2538,11 +2697,8 @@ func callHierarchy(mark marker, src protocol.Location, getCalls callHierarchyFun
 		mark.errorf("call hierarchy failed: %v", err)
 		return
 	}
-	if calls == nil {
-		calls = []protocol.Location{} // non-nil; cmp.Diff cares
-	}
-	if d := cmp.Diff(want, calls); d != "" {
-		mark.errorf("call hierarchy: unexpected results (-want +got):\n%s", d)
+	if err := compareLocations(mark, calls, want); err != nil {
+		mark.errorf("%s failed: %v", mark.note.Name, err)
 	}
 }
 
